/************************************************************************
* Name:         Measurement of FFT Execution Times
* Description:  The execution times of the fast Fourier transform (FFT) 
                can differ widely for similar sampling sizes. By means
                of this module the FFT execution times for quadratic 
                fields with different sampling numbers in float and/or 
                double precision can be measured and plotted.
* Authors:      Tino Untermann, LightTrans GmbH
* Keywords:     fast fourier transform, execution time, computing time, 
                design, module, IFTA, FFT
* Version:      1.0 (February 04, 2011)
*************************************************************************/

using System;
using System.Drawing;
using System.Collections.Generic;

using VirtualLabAPI.Core.Common;
using VirtualLabAPI.Core.Numerics;
using VirtualLabAPI.Core.FieldRepresentations;
using VirtualLabAPI.Core.Modules;
using VirtualLabAPI.Core.DataVisualization;

public class VLModule : IVLModule {
    
    // ********************* adjust module parameters ************************
    // ***********************************************************************
    // start value of field sampling in x-direction
    // (and y-direction for 2D fields)
    int _startSamplingCount = 1000;

    // start value of field sampling in x-direction
    // (and y-direction for 2D fields)
    int _endSamplingCount = 1100;
    
    // step size for increment
    int _stepSize = 1;

    // number of fft tests per sampling
    // the resulting runtime yields from the arithmetic mean
    // (execution times may vary from run to run at constant sampling)
    int _noOfRunsPerSampling = 1;

    // if true, fft runtimes for fields in float precision are measured
    bool _evaluateFloatPrecision = true;

    // if true, fft runtimes for fields in double precision are measured 
    bool _evaluateDoublePrecision = true;
    // ***********************************************************************
    // ***********************************************************************

    public void Run() {
        DateTime dtStart = new DateTime();
        DateTime dtEnd = new DateTime();
        TimeSpan timeSpan;

        double dNoRunsPerSampling = (double)_noOfRunsPerSampling;
        int noOfSamplings =
            (_endSamplingCount - _startSamplingCount) / _stepSize + 1;

        // arrays for measurement data
        double[] samplingCounts = new double[noOfSamplings];
        double[] fftFloatCompTimes = null;
        double[] fftDoubleCompTimes = null;
        if (_evaluateFloatPrecision) {
            fftFloatCompTimes = new double[noOfSamplings];
        }
        if (_evaluateDoublePrecision) {
            fftDoubleCompTimes = new double[noOfSamplings];
        }

        // run through all samplings
        for (int samplingCount = _startSamplingCount, measurementNo = 0;
             samplingCount <= _endSamplingCount;
             samplingCount += _stepSize, measurementNo++) {

             // log progress
             if(measurementNo % 10 == 0){
                Globals.DataDisplay.LogMessage("Step number " + measurementNo.ToString() + ", samplingCount=" + samplingCount.ToString());
             }

            ComplexAmplitude caFloat = null;
            ComplexAmplitude caDouble = null;
            if (_evaluateFloatPrecision) {
                caFloat = generateTopHat(false, samplingCount);
            }
            if (_evaluateDoublePrecision) {
                caDouble = generateTopHat(true, samplingCount);
            }

            double fftTime_Float = 0.0; // average fft execution time (float precision)
            double fftTime_Double = 0.0; // average fft execution time (double precision)

            for (int i = 0; i < _noOfRunsPerSampling; ++i) {

                if (_evaluateFloatPrecision) {
                    ComplexAmplitude caCopy = new ComplexAmplitude(caFloat); // clone field
                    
                    //**************** measurement of fft execution time ******************
                    dtStart = DateTime.Now;
                    // perform fast Fourier transform
                    VirtualLabAPI.Core.FieldRepresentations.Transformations.Fourier(caCopy);
                    dtEnd = DateTime.Now;
                    timeSpan = new TimeSpan(dtEnd.Ticks - dtStart.Ticks); // milliseconds
                    //*********************************************************************

                    // sum of execution times at current sampling
                    fftTime_Float += timeSpan.TotalMilliseconds;
                }
                if (_evaluateDoublePrecision) {
                    ComplexAmplitude caCopy = new ComplexAmplitude(caDouble); // clone field
                    
                    //**************** measurement of fft execution time ******************
                    dtStart = DateTime.Now;
                    // perform fast Fourier transform
                    VirtualLabAPI.Core.FieldRepresentations.Transformations.Fourier(caCopy);
                    dtEnd = DateTime.Now;
                    timeSpan = new TimeSpan(dtEnd.Ticks - dtStart.Ticks); // milliseconds
                    //*********************************************************************

                    // sum of execution times at current sampling
                    fftTime_Double += timeSpan.TotalMilliseconds;
                }
            }
            if (_evaluateFloatPrecision) {
                // calculate average time for current sampling (float precision)
                fftTime_Float /= dNoRunsPerSampling;
                // add execution times to array (in seconds)
                fftFloatCompTimes[measurementNo] = fftTime_Float / 1000;
            }
            if (_evaluateDoublePrecision) {
                // calculate average time for current sampling (double precision)
                fftTime_Double /= dNoRunsPerSampling;
                // add execution times to array (in seconds) 
                fftDoubleCompTimes[measurementNo] = fftTime_Double / 1000;
            }
            samplingCounts[measurementNo] = (double)samplingCount;

            // manual garbage collection
            GC.Collect();
            GC.WaitForPendingFinalizers();
        }

        createAndShowDiagram(samplingCounts, fftFloatCompTimes, fftDoubleCompTimes);
        Globals.DataDisplay.LogMessage("Test of fft execution times finished.");
    }

    /// <summary>
    /// Create and display a data series diagram plotting the computing times 
    /// over sampling counts
    /// </summary>
    /// <param name="samplingCounts">
    /// Array of sampling counts.</param>
    /// <param name="fftFloatCompTimes">
    /// Array of fft computing times (float precision). Can be null.</param>
    /// <param name="fftDoubleCompTimes">
    /// Array of fft computing times (double precision). Can be null.</param>
    void createAndShowDiagram(double[] samplingCounts,
                              double[] fftFloatCompTimes,
                              double[] fftDoubleCompTimes) {

        // create new data series diagram
        DataSeriesDiagramData diagram = new DataSeriesDiagramData();
        diagram.DiagramType = DataSeriesDiagramType.LinesDiagram;
        List<Color> colors = new List<Color>(2);

        // add series for float precision to diagram
        if (fftFloatCompTimes != null) {
            DataSeries dataSeries = new DataSeries();
            List<DiagramPoint2D> diagramPointList =
                new List<DiagramPoint2D>(samplingCounts.Length);
            for (int i = 0; i < samplingCounts.Length; ++i) {
                DiagramPoint2D diagramPoint = new DiagramPoint2D(samplingCounts[i],
                                                                 fftFloatCompTimes[i],
                                                                 true);
                diagramPointList.Add(diagramPoint);
            }
            dataSeries.DataPoints = diagramPointList;
            dataSeries.Label = "Float Precision";
            diagram.DataRows.Add(dataSeries);
            colors.Add(Color.Red);
        }

        // add series for double precision to diagram
        if (fftDoubleCompTimes != null) {
            DataSeries dataSeries = new DataSeries();
            List<DiagramPoint2D> diagramPointList =
                new List<DiagramPoint2D>(samplingCounts.Length);
            for (int i = 0; i < samplingCounts.Length; ++i) {
                DiagramPoint2D diagramPoint = new DiagramPoint2D(samplingCounts[i],
                                                                 fftDoubleCompTimes[i],
                                                                 true);
                diagramPointList.Add(diagramPoint);
            }
            dataSeries.DataPoints = diagramPointList;
            dataSeries.Label = "Double Precision";
            diagram.DataRows.Add(dataSeries);
            colors.Add(Color.Blue);
        }

        if (fftFloatCompTimes != null || fftDoubleCompTimes != null) {
            // set labels and units
            diagram.LabelOfXAxis = "Number of Sampling Points (x-Direction)";
            diagram.PhysicalPropertyOfXValues = PhysicalProperty.NoUnit;
            diagram.LabelOfYAxis = "FFT Execution Time";
            diagram.PhysicalPropertyOfYValues = PhysicalProperty.Time;
            diagram.Colors = colors.ToArray();

            // show diagram
            Globals.DataDisplay.ShowDocument(diagram,
                                             "FFT Execution Times vs. Number of Sampling Points");
        }
    }

    /// <summary>
    /// Generates a quadratically sampled, E-field with shape of a 
    /// quadratic top hat. Field intensity is either zero or one at the 
    /// sampling points.
    /// </summary>
    /// <param name="doublePrecision">
    /// Flag to signal if resulting field shall be in double or float precision.
    /// True for double precision, false for float precision.</param>
    /// <param name="samplingCount">
    /// Number of sampling points in x- and y-direction</param>
    /// <returns>Generated field.</returns>
    ComplexAmplitude generateTopHat(bool doublePrecision, int samplingCount) {
        // top hat parameters
        int index1 = (int)(0.25 * ((double)samplingCount));
        int index2 = samplingCount - index1;
        
        // determine precision mode
        PrecisionMode precision = PrecisionMode.Float;
        if (doublePrecision) {
            precision = PrecisionMode.Double;
        }

        // 2-D complex field initialized with zeros
        ComplexField topHatField = new ComplexField();
        topHatField = new ComplexField(new Vector(samplingCount, samplingCount),
                                       true,
                                       precision);
        Complex cOne = new Complex(1, 0);
      
        // generate quadratic top hat with half field size diameter
        // imaginary part is constant zero
        for (int i = 0; i < samplingCount; ++i) {
            for (int j = 0; j < samplingCount; ++j) {
                if (i >= index1 && i <= index2 &&
                    j >= index1 && j <= index2) {
                        topHatField[i, j] = cOne;
                }
            }
        }

        // construct a ComplexAmplitude object (globally polarized E-field) from
        // the previously generated complex field where all E-vectors
        // point to positive x-direction 
        return new ComplexAmplitude(topHatField);
    }
}